/**
 * @file libgalago/galago-person.c Galago Person API
 *
 * @Copyright (C) 2004-2005 Christian Hammond
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include <libgalago/galago-private.h>
#include <libgalago/galago-person.h>
#include <libgalago/galago-assert.h>
#include <libgalago/galago-context.h>
#include <libgalago/galago-context-priv.h>
#include <libgalago/galago-core.h>
#include <libgalago/galago-hashtable.h>
#include <libgalago/galago-marshal.h>
#include <libgalago/galago-utils.h>
#include <stdio.h>
#include <string.h>

struct _GalagoPersonPrivate
{
	galago_bool native;
	galago_bool me;

	char *id;

	GalagoPhoto *photo;

	GalagoHashTable *accounts_table;
	GalagoHashTable *properties_table;

	GalagoList *accounts;
};

typedef struct
{
	char *username;
	const GalagoService *service;

} AccountCacheKey;

static GalagoPerson *_galago_person_new_common(const char *id,
											   galago_bool native,
											   galago_bool me,
											   const char *obj_path);
static void _galago_dbus_person_set_photo(GalagoPerson *person,
										  GalagoPhoto *photo);
static GalagoPhoto *_galago_dbus_person_get_photo(const GalagoPerson *person);
static void _galago_dbus_person_set_property(GalagoPerson *person,
											 const char *name,
                                             GalagoValue *value);
static void _galago_dbus_person_remove_property(GalagoPerson *person,
                                                const char *name);
static GalagoValue *_galago_dbus_person_get_property(
	const GalagoPerson *person, const char *name);
static GalagoAccount *_galago_dbus_person_get_account(
	const GalagoPerson *person, const GalagoService *service,
	const char *username);
static GalagoList *_galago_dbus_person_get_accounts(const GalagoPerson *person);


/**************************************************************************
 * Object/Class support
 **************************************************************************/
static void
cache_key_free(AccountCacheKey *key)
{
	if (key->username != NULL)
		free(key->username);

	free(key);
}

static void
galago_person_object_init(GalagoPerson *person)
{
	person->priv = galago_new0(GalagoPersonPrivate, 1);

	person->priv->accounts_table =
		galago_hash_table_new_full(galago_str_hash, galago_str_equal,
								   (GalagoFreeFunc)cache_key_free, NULL);
}

static void
galago_person_object_finalize(GalagoObject *object)
{
	GalagoPerson *person = (GalagoPerson *)object;

	if (person->priv->accounts != NULL)
	{
		galago_list_foreach(person->priv->accounts,
							(GalagoForEachFunc)galago_object_unref, NULL);
		galago_list_destroy(person->priv->accounts);
	}

	galago_hash_table_destroy(person->priv->accounts_table);

	if (person->priv->properties_table != NULL)
		galago_hash_table_destroy(person->priv->properties_table);

	galago_context_push(galago_object_get_context(person));
	galago_context_remove_person(person);
	galago_context_pop();

	if (person->priv->id != NULL)
		free(person->priv->id);

	free(person->priv);
}

static void
galago_person_dbus_message_append(DBusMessageIter *iter,
								  const GalagoObject *object)
{
	GalagoPerson *person = (GalagoPerson *)object;

	dbus_message_iter_append_string(iter,  galago_object_get_dbus_path(person));
	dbus_message_iter_append_string(iter,  galago_person_get_id(person));
	dbus_message_iter_append_boolean(iter, galago_person_is_me(person));
}

static void *
galago_person_dbus_message_get(DBusMessageIter *iter)
{
	GalagoPerson *person;
	char *obj_path, *id;
	galago_bool me;

	obj_path = dbus_message_iter_get_string(iter);
	dbus_message_iter_next(iter);

	id = dbus_message_iter_get_string(iter);
	dbus_message_iter_next(iter);

	me = dbus_message_iter_get_boolean(iter);

	person = _galago_person_new_common(id, FALSE, me, obj_path);

	dbus_free(obj_path);
	dbus_free(id);

	return person;
}

static void
_push_key(void *key, void *val, void *user_data)
{
	const char *name     = (const char *)key;
	GalagoValue *value   = (GalagoValue *)val;
	GalagoPerson *person = (GalagoPerson *)user_data;

	_galago_dbus_person_set_property(person, name, value);
}

static void
galago_person_dbus_push_full(GalagoObject *object)
{
	GalagoPerson *person = (GalagoPerson *)object;

	_galago_dbus_person_set_photo(person,
		galago_person_get_photo(person, FALSE));

	if (person->priv->properties_table != NULL)
	{
		galago_hash_table_foreach(person->priv->properties_table,
								  _push_key, person);
	}

	/* NOTE: Accounts are already handled by this point. */
}

static void
galago_person_class_init(GalagoObjectClass *klass)
{
	klass->finalize            = galago_person_object_finalize;
	klass->dbus_message_append = galago_person_dbus_message_append;
	klass->dbus_message_get    = galago_person_dbus_message_get;
	klass->dbus_push_full      = galago_person_dbus_push_full;

	galago_signal_register(klass->signal_context, "photo-set",
						   galago_marshal_VOID__POINTER, 1,
						   galago_value_new(GALAGO_TYPE_SUBTYPE,
											GALAGO_SUBTYPE_PHOTO));
	galago_signal_register(klass->signal_context, "updated",
						   galago_marshal_VOID, 0);
}


/**************************************************************************
 * GalagoPerson API
 **************************************************************************/
static GalagoPerson *
_galago_person_new_common(const char *id, galago_bool native,
						  galago_bool me, const char *obj_path)
{
	GalagoPerson *person;

	galago_return_val_if_fail(galago_is_initted(),       NULL);
	galago_return_val_if_fail(id != NULL && *id != '\0', NULL);

	person = galago_context_get_person(id, native);

	if (person == NULL)
	{
		const char *obj_prefix;

		person = galago_object_new(GALAGO_CLASS_PERSON);

		person->priv->id     = strdup(id);
		person->priv->native = native;
		person->priv->me     = me;

		if (obj_path != NULL)
		{
			galago_object_set_dbus_path(person, obj_path);
		}
		else if ((obj_prefix = galago_context_get_obj_path_prefix()) != NULL)
		{
			char *new_obj_path;
			size_t len;

			len = strlen(obj_prefix) + strlen("/people/") + strlen(id) + 1;

			new_obj_path = galago_new(char, len);
			snprintf(new_obj_path, len, "%s/people/%s", obj_prefix, id);

			galago_object_set_dbus_path(person, new_obj_path);

			free(new_obj_path);
		}

		galago_context_add_person(person);
	}

	return person;
}

GalagoPerson *
galago_person_new(const char *id, galago_bool native)
{
	galago_return_val_if_fail(galago_is_initted(),       NULL);
	galago_return_val_if_fail(id != NULL && *id != '\0', NULL);

	return _galago_person_new_common(id, native, FALSE, NULL);
}

GalagoPerson *
galago_person_me_new(galago_bool native)
{
	return _galago_person_new_common(GALAGO_ME_ID, native, TRUE, NULL);
}

void
galago_person_set_me(GalagoPerson *person)
{
	galago_return_if_fail(person != NULL);
	galago_return_if_fail(GALAGO_IS_PERSON(person));

	if (galago_person_is_me(person))
		return;

	person->priv->me = TRUE;

	free(person->priv->id);

	person->priv->id = strdup(GALAGO_ME_ID);
}

galago_bool
galago_person_is_me(const GalagoPerson *person)
{
	galago_return_val_if_fail(person != NULL,           FALSE);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), FALSE);

	return person->priv->me;
}

const char *
galago_person_get_id(const GalagoPerson *person)
{
	galago_return_val_if_fail(person != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), NULL);

	return person->priv->id;
}

galago_bool
galago_person_is_native(const GalagoPerson *person)
{
	galago_return_val_if_fail(person != NULL,           FALSE);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), FALSE);

	return person->priv->native;
}

const char *
galago_person_get_display_name(const GalagoPerson *person)
{
	static char buffer[BUFSIZ];
	const char *first_name, *last_name;

	galago_return_val_if_fail(person != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), NULL);

	/* XXX This is all a hack, but for now, it should work. Improve this! */

	first_name = galago_person_get_property_string(person,
		GALAGO_PERSON_PROP_FIRST_NAME);
	last_name = galago_person_get_property_string(person,
		GALAGO_PERSON_PROP_LAST_NAME);

	if (first_name != NULL || last_name != NULL)
	{
		snprintf(buffer, sizeof(buffer), "%s%s%s",
				 (first_name == NULL ? "" : first_name),
				 (first_name != NULL && last_name != NULL ? " " : ""),
				 (last_name == NULL ? "" : last_name));
	}
	else
	{
		GalagoAccount *account;

		account = galago_person_get_priority_account(person);

		if (account == NULL)
			return NULL; /* Oh well. */

		strncpy(buffer, galago_account_get_display_name(account),
				sizeof(buffer));
	}

	return buffer;
}

void
galago_person_set_photo(GalagoPerson *person, GalagoPhoto *photo)
{
	galago_return_if_fail(person != NULL);
	galago_return_if_fail(GALAGO_IS_PERSON(person));
	galago_return_if_fail(photo == NULL || GALAGO_IS_PHOTO(photo));

	if (person->priv->photo == photo)
		return;

	if (person->priv->photo != NULL)
	{
		GalagoPhoto *old_photo = person->priv->photo;

		person->priv->photo = NULL;

		galago_object_unref(old_photo);
	}

	person->priv->photo = photo;

	if (galago_person_is_native(person))
		_galago_dbus_person_set_photo(person, photo);

	galago_signal_emit(person, "photo-set", photo);
	galago_signal_emit(person, "updated");
}

GalagoPhoto *
galago_person_get_photo(const GalagoPerson *person, galago_bool query)
{
	galago_return_val_if_fail(person != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), NULL);

	if (person->priv->photo == NULL && query &&
		!galago_person_is_native(person) && !galago_is_daemon())
	{
		GalagoSignalContext *signal_context = GALAGO_SIGNAL_CONTEXT(person);

		galago_signal_context_freeze(signal_context);
		person->priv->photo = _galago_dbus_person_get_photo(person);
		galago_signal_context_thaw(signal_context);
	}

	return person->priv->photo;
}

galago_bool
galago_person_has_accounts(const GalagoPerson *person, galago_bool query)
{
	galago_return_val_if_fail(person != NULL,           FALSE);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), FALSE);

	return (galago_person_get_accounts(person, query) != NULL);
}

const GalagoList *
galago_person_get_accounts(const GalagoPerson *person, galago_bool query)
{
	galago_return_val_if_fail(person != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), NULL);

	if (query && !galago_person_is_native(person) && !galago_is_daemon())
	{
		GalagoList *temp;

		galago_context_push(galago_object_get_context(person));
		/* galago_signal_context_freeze(signal_context); */

		temp = _galago_dbus_person_get_accounts(person);
		galago_list_destroy(temp);

		/* galago_signal_context_thaw(signal_context); */
		galago_context_pop();
	}

	return person->priv->accounts;
}

GalagoAccount *
galago_person_get_priority_account(const GalagoPerson *person)
{
	const GalagoList *l, *accounts;
	GalagoPresence *priority_presence = NULL;
	GalagoAccount *account = NULL;

	galago_return_val_if_fail(person != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person), NULL);

	for (l = accounts = galago_person_get_accounts(person, TRUE);
		 l != NULL;
		 l = l->next)
	{
		GalagoAccount *account = (GalagoAccount *)l->data;
		GalagoPresence *presence = galago_account_get_presence(account, TRUE);

		if (galago_presence_compare(priority_presence, presence) > 0)
			priority_presence = presence;
	}

	if (priority_presence != NULL)
	{
		account = galago_presence_get_account(priority_presence);
	}
	else if (accounts != NULL)
	{
		account = (GalagoAccount *)accounts->data;
	}

	return account;
}

void
galago_person_add_account(GalagoPerson *person, GalagoAccount *account)
{
	const char *username;
	GalagoService *service;
	AccountCacheKey *key;

	galago_return_if_fail(person  != NULL);
	galago_return_if_fail(account != NULL);
	galago_return_if_fail(GALAGO_IS_PERSON(person));
	galago_return_if_fail(GALAGO_IS_ACCOUNT(account));

	username = galago_account_get_username(account);
	service  = galago_account_get_service(account);

	if (galago_person_get_account(person, service, username, FALSE) != NULL)
	{
#if 0
		galago_log_warning("An account with username %s on service %s has "
						   "already been added to person %s\n",
						   username, galago_service_get_id(service),
						   galago_person_get_id(person));
#endif

		return;
	}

	key = galago_new0(AccountCacheKey, 1);
	key->username = strdup(username);
	key->service  = service;

	galago_hash_table_insert(person->priv->accounts_table, key, account);

	person->priv->accounts = galago_list_append(person->priv->accounts,
												account);

#if 0
	galago_signal_emit(service, "account-added", account);
	galago_signal_emit(service, "updated");
#endif
}

void
galago_person_remove_account(GalagoPerson *person, GalagoAccount *account)
{
	AccountCacheKey key;
	GalagoService *service;
	const char *username;

	galago_return_if_fail(person  != NULL);
	galago_return_if_fail(account != NULL);
	galago_return_if_fail(GALAGO_IS_PERSON(person));
	galago_return_if_fail(GALAGO_IS_ACCOUNT(account));

	service  = galago_account_get_service(account);
	username = galago_account_get_username(account);

	key.username = strdup(galago_service_normalize(service, username));
	key.service  = galago_account_get_service(account);

	galago_hash_table_remove(person->priv->accounts_table, &key);

	free(key.username);

	person->priv->accounts = galago_list_remove(person->priv->accounts,
												account);

#if 0
	galago_signal_emit(person, "account-removed", account);
	galago_signal_emit(person, "updated");
#endif
}

GalagoAccount *
galago_person_get_account(const GalagoPerson *person,
						  const GalagoService *service, const char *username,
						  galago_bool query)
{
	GalagoAccount *account;
	AccountCacheKey key;

	galago_return_val_if_fail(person   != NULL,           NULL);
	galago_return_val_if_fail(service  != NULL,           NULL);
	galago_return_val_if_fail(username != NULL,           NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),   NULL);
	galago_return_val_if_fail(GALAGO_IS_SERVICE(service), NULL);
	galago_return_val_if_fail(galago_person_is_native(person) ==
							  galago_service_is_native(service), NULL);

	key.username = strdup(galago_service_normalize(service, username));
	key.service  = service;

	account = galago_hash_table_lookup(person->priv->accounts_table, &key);

	free(key.username);

	if (account == NULL && query &&
		!galago_person_is_native(person) && !galago_is_daemon())
	{
		galago_context_push(galago_object_get_context(person));
		/* galago_signal_context_freeze(signal_context); */

		account = _galago_dbus_person_get_account(person, service, username);

		/* galago_signal_context_thaw(signal_context); */
		galago_context_pop();
	}

	return account;
}

void
galago_person_set_property_string(GalagoPerson *person, const char *name,
								  const char *str_value)
{
	GalagoValue *value;

	galago_return_if_fail(person    != NULL);
	galago_return_if_fail(name      != NULL && *name      != '\0');
	galago_return_if_fail(str_value != NULL && *str_value != '\0');
	galago_return_if_fail(GALAGO_IS_PERSON(person));

	value = galago_value_new(GALAGO_TYPE_STRING);
	galago_value_set_string(value, str_value);

	galago_person_set_property(person, name, value);
}

void
galago_person_set_property_bool(GalagoPerson *person, const char *name,
								galago_bool bool_value)
{
	GalagoValue *value;

	galago_return_if_fail(person != NULL);
	galago_return_if_fail(name   != NULL && *name  != '\0');
	galago_return_if_fail(GALAGO_IS_PERSON(person));

	value = galago_value_new(GALAGO_TYPE_BOOLEAN);
	galago_value_set_boolean(value, bool_value);

	galago_person_set_property(person, name, value);
}

void
galago_person_set_property_uint32(GalagoPerson *person, const char *name,
								  dbus_uint32_t uint_value)
{
	GalagoValue *value;

	galago_return_if_fail(person != NULL);
	galago_return_if_fail(name   != NULL && *name  != '\0');
	galago_return_if_fail(GALAGO_IS_PERSON(person));

	value = galago_value_new(GALAGO_TYPE_UINT);
	galago_value_set_uint(value, uint_value);

	galago_person_set_property(person, name, value);
}

#define IS_TYPE_INTEGER(type) \
	((type) == GALAGO_TYPE_UINT    || \
	 (type) == GALAGO_TYPE_INT     || \
	 (type) == GALAGO_TYPE_SHORT   || \
	 (type) == GALAGO_TYPE_USHORT  || \
	 (type) == GALAGO_TYPE_LONG    || \
	 (type) == GALAGO_TYPE_ULONG)

void
galago_person_set_property(GalagoPerson *person, const char *name,
						   GalagoValue *value)
{
	GalagoType value_type;

	galago_return_if_fail(person != NULL);
	galago_return_if_fail(name   != NULL && *name  != '\0');
	galago_return_if_fail(value  != NULL);
	galago_return_if_fail(GALAGO_IS_PERSON(person));

	value_type = galago_value_get_type(value);

	galago_return_if_fail(value_type == GALAGO_TYPE_STRING  ||
						  value_type == GALAGO_TYPE_BOOLEAN ||
						  IS_TYPE_INTEGER(value_type));

	if (person->priv->properties_table == NULL)
	{
		person->priv->properties_table =
			galago_hash_table_new_full(galago_str_hash, galago_str_equal,
									   free,
									   (GalagoFreeFunc)galago_value_destroy);
	}

	galago_hash_table_replace(person->priv->properties_table,
							  galago_str_lower(name), value);

	if (galago_person_is_native(person))
		_galago_dbus_person_set_property(person, name, value);
}

galago_bool
galago_person_remove_property(GalagoPerson *person, const char *name)
{
	char *temp;

	galago_return_val_if_fail(person != NULL,                  FALSE);
	galago_return_val_if_fail(name   != NULL && *name != '\0', FALSE);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),        FALSE);

	if (person->priv->properties_table == NULL)
		return TRUE;

	temp = galago_str_lower(name);
	galago_hash_table_remove(person->priv->properties_table, temp);
	free(temp);

	if (galago_person_is_native(person))
		_galago_dbus_person_remove_property(person, name);

	return TRUE;
}

const char *
galago_person_get_property_string(const GalagoPerson *person,
								  const char *name)
{
	const GalagoValue *value;

	galago_return_val_if_fail(person != NULL,                  NULL);
	galago_return_val_if_fail(name   != NULL && *name != '\0', NULL);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),        NULL);

	value = galago_person_get_property(person, name);

	if (value == NULL)
		return NULL;

	galago_return_val_if_fail(
		galago_value_get_type(value) == GALAGO_TYPE_STRING, NULL);

	return galago_value_get_string(value);
}

galago_bool
galago_person_get_property_bool(const GalagoPerson *person, const char *name)
{
	const GalagoValue *value;

	galago_return_val_if_fail(person != NULL,                  FALSE);
	galago_return_val_if_fail(name   != NULL && *name != '\0', FALSE);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),        FALSE);

	value = galago_person_get_property(person, name);

	if (value == NULL)
		return FALSE;

	galago_return_val_if_fail(
		galago_value_get_type(value) == GALAGO_TYPE_BOOLEAN, FALSE);

	return galago_value_get_boolean(value);
}

dbus_uint32_t
galago_person_get_property_uint32(const GalagoPerson *person,
								  const char *name)
{
	const GalagoValue *value;

	galago_return_val_if_fail(person != NULL,                  0);
	galago_return_val_if_fail(name   != NULL && *name != '\0', 0);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),        0);

	value = galago_person_get_property(person, name);

	if (value == NULL)
		return 0;

	galago_return_val_if_fail(
		galago_value_get_type(value) == GALAGO_TYPE_UINT, 0);

	return galago_value_get_uint(value);
}

const GalagoValue *
galago_person_get_property(const GalagoPerson *person, const char *name)
{
	GalagoValue *value = NULL;
	char *temp;

	galago_return_val_if_fail(person != NULL,                  0);
	galago_return_val_if_fail(name   != NULL && *name != '\0', 0);
	galago_return_val_if_fail(GALAGO_IS_PERSON(person),        0);

	temp = galago_str_lower(name);

	if (person->priv->properties_table != NULL)
		value = galago_hash_table_lookup(person->priv->properties_table, temp);

	if (value == NULL && !galago_person_is_native(person))
		value = _galago_dbus_person_get_property(person, temp);

	free(temp);

	return value;
}

/**************************************************************************
 * People Subsystem API
 **************************************************************************/
GalagoObjectClass *
galago_people_get_class(void)
{
	static GalagoObjectClass *klass = NULL;

	if (klass == NULL)
	{
		static const GalagoObjectClassInfo info =
		{
			0,
			sizeof(GalagoPersonClass),
			galago_person_class_init,
			NULL,
			sizeof(GalagoPerson),
			(GalagoObjectInitFunc)galago_person_object_init,
		};

		klass = galago_class_register(NULL, "GalagoPerson",
									  GALAGO_DBUS_PERSON_INTERFACE,
									  &info);
	}

	return klass;
}


/**************************************************************************
 * D-BUS Functions
 **************************************************************************/
static void
_galago_dbus_person_set_photo(GalagoPerson *person, GalagoPhoto *photo)
{
	DBusConnection *dbus_conn;
	DBusMessage *message;
	DBusMessageIter iter;

	if (!galago_is_connected() || !galago_core_is_feed())
		return;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "SetPhoto",
												  FALSE, &iter);

	galago_return_if_fail(message != NULL);

	galago_dbus_message_iter_append_object(&iter, photo);

	dbus_connection_send(dbus_conn, message, NULL);
	dbus_message_unref(message);
}

static GalagoPhoto *
_galago_dbus_person_get_photo(const GalagoPerson *person)
{
	DBusConnection *dbus_conn;
	DBusMessage *message, *reply;
	DBusMessageIter iter;
	DBusError error;
	GalagoPhoto *photo;

	if (!galago_is_connected())
		return NULL;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "GetPhoto",
												  TRUE, NULL);

	galago_return_val_if_fail(message != NULL, NULL);

	dbus_error_init(&error);

	reply = dbus_connection_send_with_reply_and_block(dbus_conn, message,
													  -1, &error);

	dbus_message_unref(message);

	if (dbus_error_is_set(&error))
	{
		galago_log_error("Error sending person.GetPhoto(%s): %s\n",
						 galago_person_get_id(person), error.message);
		dbus_error_free(&error);

		return NULL;
	}

	dbus_message_iter_init(reply, &iter);
	photo = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_PHOTO);

	dbus_message_unref(reply);
	dbus_error_free(&error);

	return photo;
}

static void
_galago_dbus_person_set_property(GalagoPerson *person, const char *name,
								 GalagoValue *value)
{
	DBusConnection *dbus_conn;
	DBusMessage *message;
	DBusMessageIter iter;
	GalagoType type;

	if (!galago_is_connected() || !galago_core_is_feed())
		return;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "SetProperty",
												  FALSE, &iter);

	galago_return_if_fail(message != NULL);

	dbus_message_iter_append_string(&iter, name);

	type = galago_value_get_type(value);

	if (type == GALAGO_TYPE_STRING)
	{
		dbus_message_iter_append_string(&iter, galago_value_get_string(value));
	}
	else if (type == GALAGO_TYPE_BOOLEAN)
	{
		dbus_message_iter_append_boolean(&iter,
										 galago_value_get_boolean(value));
	}
	else if (type == GALAGO_TYPE_UINT)
	{
		dbus_message_iter_append_uint32(&iter, galago_value_get_uint(value));
	}
	else
	{
		galago_log_error("Unknown property type %d for property %s\n",
						 type, name);

		dbus_message_unref(message);

		return;
	}

	dbus_connection_send(dbus_conn, message, NULL);
	dbus_message_unref(message);
}

static void
_galago_dbus_person_remove_property(GalagoPerson *person, const char *name)
{
	DBusConnection *dbus_conn;
	DBusMessage *message;
	DBusMessageIter iter;

	if (!galago_is_connected() || !galago_core_is_feed())
		return;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "RemoveProperty",
												  FALSE, &iter);

	galago_return_if_fail(message != NULL);

	dbus_message_iter_append_string(&iter, name);

	dbus_connection_send(dbus_conn, message, NULL);
	dbus_message_unref(message);
}

static GalagoValue *
_galago_dbus_person_get_property(const GalagoPerson *person, const char *name)
{
	DBusConnection *dbus_conn;
	DBusMessage *message, *reply;
	DBusMessageIter iter;
	DBusError error;
	GalagoValue *value = NULL;

	if (!galago_is_connected())
		return NULL;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "GetProperty",
												  TRUE, &iter);

	galago_return_val_if_fail(message != NULL, NULL);

	dbus_message_iter_append_string(&iter, name);

	dbus_error_init(&error);

	reply = dbus_connection_send_with_reply_and_block(dbus_conn, message,
													  -1, &error);

	dbus_message_unref(message);

	if (dbus_error_is_set(&error))
	{
		galago_log_error("Error sending Person.GetProperty(%s, %s): %s\n",
						 galago_person_get_id(person), name,
						 error.message);
		dbus_error_free(&error);

		return NULL;
	}

	dbus_message_iter_init(reply, &iter);

	int type = dbus_message_iter_get_arg_type(&iter);

	if (type == DBUS_TYPE_STRING)
	{
		value = galago_value_new(GALAGO_TYPE_STRING);
		galago_value_set_string(value, dbus_message_iter_get_string(&iter));
	}
	else if (type == DBUS_TYPE_BOOLEAN)
	{
		value = galago_value_new(GALAGO_TYPE_BOOLEAN);
		galago_value_set_boolean(value, dbus_message_iter_get_boolean(&iter));
	}
	else if (type == DBUS_TYPE_UINT32)
	{
		value = galago_value_new(GALAGO_TYPE_UINT);
		galago_value_set_uint(value, dbus_message_iter_get_uint32(&iter));
	}
	else if (type != DBUS_TYPE_NIL)
	{
		galago_log_fatal("Unknown property type %d for property %s\n",
						 type, name);
	}

	dbus_message_unref(reply);
	dbus_error_free(&error);

	return value;
}

static GalagoAccount *
_galago_dbus_person_get_account(const GalagoPerson *person,
								const GalagoService *service,
								const char *username)
{
	DBusConnection *dbus_conn;
	DBusMessage *message, *reply;
	DBusMessageIter iter;
	DBusError error;
	GalagoAccount *account;

	if (!galago_is_connected())
		return NULL;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "GetAccount",
												  TRUE, &iter);

	galago_return_val_if_fail(message != NULL, NULL);

	galago_dbus_message_iter_append_object(&iter, service);
	dbus_message_iter_append_string(&iter, username);

	dbus_error_init(&error);

	reply = dbus_connection_send_with_reply_and_block(dbus_conn, message,
													  -1, &error);

	dbus_message_unref(message);

	if (dbus_error_is_set(&error))
	{
		galago_log_error("Error sending Person.GetAccount(%s, %s, %s): %s\n",
						 galago_person_get_id(person),
						 galago_service_get_id(service), username,
						 error.message);
		dbus_error_free(&error);

		return NULL;
	}

	dbus_message_iter_init(reply, &iter);
	account = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_ACCOUNT);

	dbus_message_unref(reply);
	dbus_error_free(&error);

	return account;
}

static GalagoList *
_galago_dbus_person_get_accounts(const GalagoPerson *person)
{
	DBusConnection *dbus_conn;
	DBusMessage *message, *reply;
	DBusMessageIter iter;
	DBusError error;
	GalagoList *accounts;

	if (!galago_is_connected())
		return NULL;

	dbus_conn = galago_core_get_dbus_conn();

	message = galago_dbus_message_new_method_call(person, "GetAccounts",
												  TRUE, NULL);

	galago_return_val_if_fail(message != NULL, NULL);

	dbus_error_init(&error);

	reply = dbus_connection_send_with_reply_and_block(dbus_conn, message,
													  -1, &error);

	dbus_message_unref(message);

	if (dbus_error_is_set(&error))
	{
		galago_log_error("Error sending Person.GetAccounts(%s): %s\n",
						 galago_person_get_id(person), error.message);
		dbus_error_free(&error);

		return NULL;
	}

	dbus_message_iter_init(reply, &iter);
	accounts = galago_dbus_message_iter_get_object_list(&iter,
														GALAGO_CLASS_ACCOUNT);

	dbus_message_unref(reply);
	dbus_error_free(&error);

	return accounts;
}
