/**
 * @file libgalago/galago-value.c Value wrapper API
 *
 * @Copyright (C) 2004-2005 Christian Hammond
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include <libgalago/galago-value.h>
#include <libgalago/galago-assert.h>
#include <libgalago/galago-utils.h>
#include <string.h>

#define OUTGOING_FLAG 0x01

#define DEFINE_ACCESSOR(name, type, defaultval) \
void \
galago_value_set_##name(GalagoValue *value, type data) \
{ \
	galago_return_if_fail(value != NULL); \
\
	value->data.name##_data = data; \
} \
\
type \
galago_value_get_##name(const GalagoValue *value) \
{ \
	galago_return_val_if_fail(value != NULL, defaultval); \
\
	return value->data.name##_data; \
}

#define DEFINE_ACCESSOR_MEM(name, type, defaultval, destroy, cpy) \
void \
galago_value_set_##name(GalagoValue *value, type data) \
{ \
	galago_return_if_fail(value != NULL); \
\
	if (value->data.name##_data != NULL) \
		(destroy)(value->data.name##_data); \
\
	value->data.name##_data = (data == NULL ? NULL : (cpy)(data)); \
} \
\
type \
galago_value_get_##name(const GalagoValue *value) \
{ \
	galago_return_val_if_fail(value != NULL, defaultval); \
\
	return value->data.name##_data; \
}

GalagoValue *
galago_value_new(GalagoType type, ...)
{
	GalagoValue *value;
	va_list args;

	galago_return_val_if_fail(type != GALAGO_TYPE_UNKNOWN, NULL);

	value = galago_new0(GalagoValue, 1);

	value->type = type;

	va_start(args, type);

	if (type == GALAGO_TYPE_SUBTYPE)
		value->u.subtype = va_arg(args, int);
	else if (type == GALAGO_TYPE_BOXED)
		value->u.specific_type = strdup(va_arg(args, char *));

	va_end(args);

	return value;
}

GalagoValue *
galago_value_new_outgoing(GalagoType type, ...)
{
	GalagoValue *value;
	va_list args;

	galago_return_val_if_fail(type != GALAGO_TYPE_UNKNOWN, NULL);

	value = galago_new0(GalagoValue, 1);

	value->type = type;

	va_start(args, type);

	if (type == GALAGO_TYPE_SUBTYPE)
		value->u.subtype = va_arg(args, int);
	else if (type == GALAGO_TYPE_BOXED)
		value->u.specific_type = strdup(va_arg(args, char *));

	va_end(args);

	value->flags |= OUTGOING_FLAG;

	return value;
}

void
galago_value_destroy(GalagoValue *value)
{
	galago_return_if_fail(value != NULL);

	if (galago_value_get_type(value) == GALAGO_TYPE_BOXED)
	{
		if (value->u.specific_type != NULL)
			free(value->u.specific_type);
	}
	else if (galago_value_get_type(value) == GALAGO_TYPE_STRING)
	{
		if (value->data.string_data != NULL)
			free(value->data.string_data);
	}

	free(value);
}

GalagoType
galago_value_get_type(const GalagoValue *value)
{
	galago_return_val_if_fail(value != NULL, GALAGO_TYPE_UNKNOWN);

	return value->type;
}

unsigned int
galago_value_get_subtype(const GalagoValue *value)
{
	galago_return_val_if_fail(value != NULL, GALAGO_SUBTYPE_UNKNOWN);
	galago_return_val_if_fail(galago_value_get_type(value) ==
							  GALAGO_TYPE_SUBTYPE, GALAGO_SUBTYPE_UNKNOWN);

	return value->u.subtype;
}

const char *
galago_value_get_specific_type(const GalagoValue *value)
{
	galago_return_val_if_fail(value != NULL, NULL);
	galago_return_val_if_fail(galago_value_get_type(value) ==
							  GALAGO_TYPE_BOXED, NULL);

	return value->u.specific_type;
}

galago_bool
galago_value_is_outgoing(const GalagoValue *value)
{
	galago_return_val_if_fail(value != NULL, FALSE);

	return (value->flags & OUTGOING_FLAG);
}

DEFINE_ACCESSOR(char,       char,           '\0')
DEFINE_ACCESSOR(uchar,      unsigned char,  '\0')
DEFINE_ACCESSOR(boolean,    galago_bool,    FALSE)
DEFINE_ACCESSOR(short,      short,          0)
DEFINE_ACCESSOR(ushort,     unsigned short, 0)
DEFINE_ACCESSOR(int,        int,            0)
DEFINE_ACCESSOR(uint,       unsigned int,   0)
DEFINE_ACCESSOR(long,       long,           0)
DEFINE_ACCESSOR(ulong,      unsigned long,  0)
DEFINE_ACCESSOR(object,     void *,         NULL)
DEFINE_ACCESSOR(pointer,    void *,         NULL)
DEFINE_ACCESSOR(enum,       int,            -1)
DEFINE_ACCESSOR(boxed,      void *,         NULL)
DEFINE_ACCESSOR_MEM(string, const char *,   NULL, free, strdup)
